#include <iostream>
#include <cstdio>
#include <thread>
#include <mutex>

using namespace std;

#define BUFFER_DIM 1024
#define BUFFER_SIZE 64
struct BUFFER
{
 unsigned char buffer[BUFFER_DIM];
 int n;
};

BUFFER ring_buffer[BUFFER_SIZE];
bool eof = false;
int read_index = 0;
int write_index = 0;
int N_block = 0;
mutex critical;

// lettura dei dati da file
void leggi(FILE* file)
{
 while (!feof(file))
      {
       // attesa nel caso di buffer pieno
       while (true)
            {
             critical.lock();
             if (N_block < BUFFER_SIZE)
               {
                critical.unlock();
                break;
               }
             critical.unlock();
            }
       // lettura di un blocco dati dal file nel buffer circolare
       int n = fread(ring_buffer[write_index].buffer, 1, BUFFER_DIM, file);
       if (n > 0)
         {
          // aggiornamento numero di byte del blocco
          ring_buffer[write_index].n = n;
          // incremento circolare indice di scrittura blocchi dati
          write_index = (write_index + 1) % BUFFER_SIZE;
          // incremento contatore numero di blocchi dati
          critical.lock();
          N_block++;
          critical.unlock();
         }
      }
 eof = true;
}

// scrittura dei dati su file
void scrivi(FILE* file)
{
 while (true)
      {
       critical.lock();
       if (eof && N_block == 0)
         {
          critical.unlock();
          break;
         }
       // nel caso di buffer vuoto si attende ripetendo il ciclo
       if (N_block > 0)
         {
          critical.unlock();
          // scrittura su file di un blocco dati del buffer circolare
          fwrite( ring_buffer[read_index].buffer, 1, ring_buffer[read_index].n, file);
          // decremento circolare indice di lettura blocchi dati
          read_index = (read_index + 1) % BUFFER_SIZE;
          // decremento contatore numero di blocchi dati
          critical.lock();
          N_block--;
          critical.unlock();
         }
       else
           critical.unlock();
      }
}


int main(int argc, char*argv[])
{
 FILE *origine, *destinazione;

 // verifica degli argomenti della riga di comando
 if (argc != 3)
   {
    cout << "Uso: " << argv[0] << " file-origine file-destinazione" << endl;
    return -1;
   }
 // apertura in lettura del file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    cout << "Errore apertura file" << argv[1] << endl;
    return -1;
   }
 destinazione = fopen(argv[2], "wb");
 // apertura in scrittura del file di destinazione
 if (destinazione == NULL)
   {
    cout << "Errore apertura file" << argv[2] << endl;
    fclose(origine);
    return -1;
   }
 // esecuzione dei thread di lettura e scrittura
 thread lettura(leggi, origine);
 thread scrittura(scrivi, destinazione);
 // attesa della conclusione dei thread e chiusura dei file
 lettura.join();
 scrittura.join();
 fclose(origine);
 fclose(destinazione);
 return 0;
}
