#include <iostream>
#include <cstdio>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <atomic>

using namespace std;

#define BUFFER_DIM 1024
#define BUFFER_SIZE 64

struct BUFFER
{
 unsigned char buffer[BUFFER_DIM];
 int n;
};

BUFFER ring_buffer[BUFFER_SIZE];
bool eof = false;
int read_index = 0;
int write_index = 0;
atomic<int> N_block;
mutex producer, consumer;
condition_variable_any not_full, not_empty;

// lettura dei dati da file
void leggi(FILE* file)
{
 while (!feof(file))
      {
       producer.lock();
       // sospensione nel caso di buffer pieno
       if (N_block >= BUFFER_SIZE)
         not_full.wait(producer);
       // lettura di un blocco dati dal file nel buffer circolare
       int n = fread(ring_buffer[write_index].buffer, 1, BUFFER_DIM, file);
       if (n > 0)
         {
          // aggiornamento numero di byte del blocco
          ring_buffer[write_index].n = n;
          // incremento circolare indice di scrittura blocchi dati
          write_index = (write_index + 1) % BUFFER_SIZE;
          // incremento contatore numero di blocchi dati
          N_block++;
		  // notifica condizione buffer non vuoto al consumatore
		  not_empty.notify_one();
         }
	   producer.unlock();
      }
 eof = true;
 // notifica al consumatore eventualmente sospeso in attesa di dati
 not_empty.notify_one();
}

// scrittura dei dati su file
void scrivi(FILE* file)
{
 while (!eof || N_block > 0)
      {
       consumer.lock();
       if (N_block > 0)
         {
          // scrittura su file di un blocco dati del buffer circolare
          fwrite( ring_buffer[read_index].buffer, 1, ring_buffer[read_index].n, file);
          // decremento circolare indice di lettura blocchi dati
          read_index = (read_index + 1) % BUFFER_SIZE;
          // decremento contatore numero di blocchi dati
          N_block--;
          // notifica condizione buffer non pieno al produttore
          not_full.notify_one();
         }
       else
           {
            // sospensione nel caso di buffer vuoto
            not_empty.wait(consumer);
           }
       consumer.unlock();
      }
}


int main(int argc, char*argv[])
{
 FILE *origine, *destinazione;

 // verifica degli argomenti della riga di comando
 if (argc != 3)
   {
    cout << "Uso: " << argv[0] << " file-origine file-destinazione" << endl;
    return -1;
   }
 // apertura in lettura del file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    cout << "Errore apertura file" << argv[1] << endl;
    return -1;
   }
 destinazione = fopen(argv[2], "wb");
 // apertura in scrittura del file di destinazione
 if (destinazione == NULL)
   {
    cout << "Errore apertura file" << argv[2] << endl;
    fclose(origine);
    return -1;
   }
 // inizializzazione variabili condivise
 N_block = 0;
 // esecuzione dei thread di lettura e scrittura
 thread lettura(leggi, origine);
 thread scrittura(scrivi, destinazione);
 // attesa della conclusione dei thread e chiusura dei file
 lettura.join();
 scrittura.join();
 fclose(origine);
 fclose(destinazione);
 return 0;
}
