
#include <stdio.h>
#include <pthread.h>

#define BUFFER_DIM 1024
#define BUFFER_SIZE 64

struct BUFFER
{
 unsigned char buffer[BUFFER_DIM];
 int n;
};

struct BUFFER ring_buffer[BUFFER_SIZE];
int read_index = 0;
int write_index = 0;
int N_block = 0;
int eof = 0;
pthread_mutex_t critical;
FILE *origine, *destinazione;


void* leggi(void* par)
{
 int n;

 while (!feof(origine))
      {
       // attesa nel caso di buffer pieno
       while (1)
            {
             pthread_mutex_lock(&critical);
             if (N_block < BUFFER_SIZE)
               {
                pthread_mutex_unlock(&critical);
                break;
               }
             pthread_mutex_unlock(&critical);
            }
       // lettura di un blocco dati dal file nel buffer circolare
       n = fread( ring_buffer[write_index].buffer, 1, BUFFER_DIM, origine);
       if (n > 0)
         {
          // aggiornamento numero di byte del blocco
          ring_buffer[write_index].n = n;
          // incremento circolare indice di scrittura blocchi dati
          write_index = (write_index + 1) % BUFFER_SIZE;
          // incremento contatore numero di blocchi dati
          pthread_mutex_lock(&critical);
          N_block++;
          pthread_mutex_unlock(&critical);
         }
      }
 eof = 1;
 pthread_exit(NULL);
}

// scrittura dei dati su file
void* scrivi(void* par)
{
 while (1)
      { 
       pthread_mutex_lock(&critical);
       if (eof && N_block == 0)
         {
          pthread_mutex_unlock(&critical);
          break;
         }
       // nel caso di buffer vuoto si attende ripetendo il ciclo
       if (N_block > 0)
         {
          pthread_mutex_unlock(&critical);
          // scrittura su file di un blocco dati del buffer circolare
          fwrite( ring_buffer[read_index].buffer, 1,
          ring_buffer[read_index].n, destinazione);
          // decremento circolare indice di lettura blocchi dati
          read_index = (read_index + 1) % BUFFER_SIZE;
          // decremento contatore numero di blocchi dati
          pthread_mutex_lock(&critical);
          N_block--;
          pthread_mutex_unlock(&critical);
         }
       else
           pthread_mutex_unlock(&critical);
      }
 pthread_exit(NULL);
}


int main(int argc, char* argv[])
{
 pthread_t thread_scrittura, thread_lettura;

 if (argc != 3)
   {
    printf("Uso: %s file-origine file-destinazione\r\n", argv[0]);
    return -1;
   }

 // apertura in lettura file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    printf ("Errore apertura file %s\r\n", argv[1]);
    return -1;
   }
 destinazione = fopen(argv[2], "wb");
 // apertura in scrittura file di destinazione
 if (destinazione == NULL)
   {
    printf("Errore apertura file %s\r\n", argv[2]);
    fclose(origine); 
    return -1;
   }
 // creazione oggetto mutex per accesso sezioni critiche
 pthread_mutex_init(&critical, NULL);
 // creazione thread di lettura e scrittura dei file
 pthread_create(&thread_lettura, NULL, &leggi, NULL);
 pthread_create(&thread_scrittura, NULL, &scrivi, NULL);
 // attesa terminazione thread
 pthread_join(thread_lettura, NULL);
 pthread_join(thread_scrittura, NULL);
 // distruzione oggetto mutex per accesso sezioni critiche
 pthread_mutex_destroy(&critical);
 // chiusura file
 fclose(origine);
 fclose(destinazione);
 return 0;
}
