
#include <stdio.h>
#include <pthread.h>
#include <stdatomic.h>

#define BUFFER_DIM 1024
#define BUFFER_SIZE 64

struct BUFFER
{
 unsigned char buffer[BUFFER_DIM];
 int n;
};

struct BUFFER ring_buffer[BUFFER_SIZE];
int read_index = 0;
int write_index = 0;
_Atomic int N_block = 0;
int eof = 0;
pthread_mutex_t producer, consumer;
pthread_cond_t not_full, not_empty;
FILE *origine, *destinazione;


void* leggi(void* par)
{
 int n;

 while (!feof(origine))
      {
       pthread_mutex_lock(&producer);
       // sospensione nel caso di buffer pieno
       if (N_block >= BUFFER_SIZE)
         pthread_cond_wait(&not_full, &producer);
       // lettura di un blocco dati dal file nel buffer circolare
       n = fread( ring_buffer[write_index].buffer, 1, BUFFER_DIM, origine);
       if (n > 0)
         {
          // aggiornamento numero di byte del blocco
          ring_buffer[write_index].n = n;
          // incremento circolare indice di scrittura blocchi dati
          write_index = (write_index + 1) % BUFFER_SIZE;
          // incremento contatore numero di blocchi dati
          N_block++;
          // notifica condizione di buffer non vuoto al consumatore
          pthread_cond_signal(&not_empty);
         }
	   pthread_mutex_unlock(&producer);
      }
 eof = 1;
 // notifica al consumatore eventualmente sospeso in attesa di dati
 pthread_cond_signal(&not_empty);
 pthread_exit(NULL);
}

// scrittura dei dati su file
void* scrivi(void* par)
{
 while (!eof || N_block > 0)
      {
       pthread_mutex_lock(&consumer);
	   if (N_block > 0)
	     {
          // scrittura su file di un blocco dati del buffer circolare
          fwrite( ring_buffer[read_index].buffer, 1, ring_buffer[read_index].n, destinazione);
          // decremento circolare indice di lettura blocchi dati
          read_index = (read_index + 1) % BUFFER_SIZE;
          // decremento contatore numero di blocchi dati
          N_block--;
          // notifica condizione buffer non pieno al produttore
          pthread_cond_signal(&not_full);
         }
       else
          {
           // sospensione nel caso di buffer vuoto
           pthread_cond_wait(&not_empty, &consumer);
          }
       pthread_mutex_unlock(&consumer)
      }
 pthread_exit(NULL);
}


int main(int argc, char* argv[])
{
 pthread_t thread_scrittura, thread_lettura;

 if (argc != 3)
   {
    printf("Uso: %s file-origine file-destinazione\r\n", argv[0]);
    return -1;
   }

 // apertura in lettura file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    printf ("Errore apertura file %s\r\n", argv[1]);
    return -1;
   }
 destinazione = fopen(argv[2], "wb");
 // apertura in scrittura file di destinazione
 if (destinazione == NULL)
   {
    printf("Errore apertura file %s\r\n", argv[2]);
    fclose(origine); 
    return -1;
   }
 // creazione oggetti mutex e variabili di condizione
 pthread_mutex_init(&producer, NULL);
 pthread_mutex_init(&consumer, NULL);
 pthread_cond_init(&not_full, NULL);
 pthread_cond_init(&not_empty, NULL);
 // creazione thread di lettura e scrittura dei file
 pthread_create(&thread_lettura, NULL, &leggi, NULL);
 pthread_create(&thread_scrittura, NULL, &scrivi, NULL);
 // attesa terminazione thread
 pthread_join(thread_lettura, NULL);
 pthread_join(thread_scrittura, NULL);
 // distruzione oggetti mutex e variabili di condizione
 pthread_mutex_destroy(&producer);
 pthread_mutex_destroy(&consumer);
 pthread_cond_destroy(&not_full);
 pthread_cond_destroy(&not_empty);
 // chiusura file
 fclose(origine);
 fclose(destinazione);
 return 0;
}
