#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <atomic>

using namespace std;

#define BUFFER_SIZE 16

int ring_buffer[BUFFER_SIZE];
bool eos = false;
int read_index = 0;
int write_index = 0;
atomic<int> N_block;
atomic<int> N_number;
mutex producer, consumer;
condition_variable_any not_full, not_empty;

// generazione numeri casuali
void genera(void)
{
 while (N_number > 0)
      {
       producer.lock();
       // sospensione nel caso di buffer pieno
       if (N_block >= BUFFER_SIZE)
         not_full.wait(producer);
       // generazione nuovo numero casuale e inserimento nel buffer circolare
       ring_buffer[write_index] = rand();
       N_number--;
       // incremento circolare indice di scrittura blocchi dati
       write_index = (write_index + 1) % BUFFER_SIZE;
       // incremento contatore numero di blocchi dati
       N_block++;
	   // notifica condizione buffer non vuoto al consumatore
	   not_empty.notify_one();
	   producer.unlock();
      }
 eos = true;
 // notifica al consumatore eventualmente sospeso in attesa di dati
 not_empty.notify_one();
}

// visualizzazione numeri casuali
void visualizza(void)
{
 while (!eos || N_block > 0)
      {
       consumer.lock();
       if (N_block > 0)
         {
          // visualizzazione numero casuale prelevato dal buffer circolare
          cout << ring_buffer[read_index] << endl;
          // decremento circolare indice di lettura blocchi dati
          read_index = (read_index + 1) % BUFFER_SIZE;
          // decremento contatore numero di blocchi dati
          N_block--;
          // notifica condizione buffer non pieno al produttore
          not_full.notify_one();
         }
       else
           {
            // sospensione nel caso di buffer vuoto
            not_empty.wait(consumer);
           }
       consumer.unlock();
      }
}


int main(int argc, char*argv[])
{
 // verifica degli argomenti della riga di comando
 if (argc != 2)
   {
    cout << "Uso: " << argv[0] << " numero-generazioni" << endl;
    return -1;
   }
 N_number = atoi(argv[1]);
 // inizializzazione variabili condivise
 N_block = 0;
 // esecuzione dei thread di lettura e scrittura
 thread generazione(genera);
 thread visualizzazione(visualizza);
 // attesa della conclusione dei thread e chiusura dei file
 generazione.join();
 visualizzazione.join();
 return 0;
}
