
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>

#include "des.h"

#define BUFFER_BLOCKS 64 // blocchi di 8 caratteri in un buffer

#define TRUE 1
#define FALSE 0

FILE *origine, *destinazione;
int eof_flag;
int crypt;
char input_buffer[BUFFER_BLOCKS*8];
int input_buffer_write_index; // indice per la scrittura nel buffer
int input_buffer_read_index; // indice per la lettura dal buffer
int input_buffer_count; // numero di blocchi nel buffer
char output_buffer[BUFFER_BLOCKS*8];
int output_buffer_write_index; // indice per la scrittura nel buffer
int output_buffer_read_index; // indice per la lettura dal buffer
int output_buffer_count; // numero di blocchi nel buffer
char password[9] = {'\0','\0','\0','\0','\0','\0','\0','\0','\0'};
// oggetti mutex per implementazione sezioni critiche di codice
pthread_mutex_t input_buffer_mutex;
pthread_mutex_t output_buffer_mutex;
// variabili condizione
pthread_cond_t input_buffer_not_full;
pthread_cond_t input_buffer_not_empty;
pthread_cond_t output_buffer_not_full;
pthread_cond_t output_buffer_not_empty;

void* read_thread(void* arg) // lettura da file
{
 int count;
 
 // lettura di un blocco di dati dal file
 fread(&input_buffer[input_buffer_write_index*8], 8, 1, origine);
 pthread_mutex_lock(&input_buffer_mutex);
 // incremento indice di scrittura nel buffer di input
 input_buffer_write_index = ( input_buffer_write_index + 1) % BUFFER_BLOCKS;
 // incremento contatore blocchi di dati nel buffer di input
 input_buffer_count++;
 pthread_mutex_unlock(&input_buffer_mutex);
 // segnalazione condizione input buffer non vuoto
 pthread_cond_signal(&input_buffer_not_empty);
 while (!feof(origine))
      {
	     // verifica/attesa disponibilita spazio per nuovi dati
       pthread_mutex_lock(&input_buffer_mutex);
       if (input_buffer_count >= BUFFER_BLOCKS) // input buffer pieno
         pthread_cond_wait( &input_buffer_not_full, &input_buffer_mutex);
       pthread_mutex_unlock(&input_buffer_mutex);
       // lettura di un blocco di dati dal file
       fread(&input_buffer[input_buffer_write_index*8], 8, 1, origine);
       pthread_mutex_lock(&input_buffer_mutex);
       // incremento indice di scrittura nel buffer di input
       input_buffer_write_index = (input_buffer_write_index + 1) % BUFFER_BLOCKS;
       // incremento contatore blocchi dati nel buffer di input
       input_buffer_count++;
       pthread_mutex_unlock(&input_buffer_mutex);
       // segnalazione condizione input buffer non vuoto
       pthread_cond_signal(&input_buffer_not_empty);
      }
 // attesa elaborazione dati presenti nel buffer di input
 do {
     pthread_mutex_lock(&input_buffer_mutex);
     count = input_buffer_count;
     pthread_mutex_unlock(&input_buffer_mutex);
     usleep(1000); // attesa 1ms
    } while (count > 0);
 // attesa scrittura dati presenti nel buffer di output
 do {
     pthread_mutex_lock(&output_buffer_mutex);
     count = output_buffer_count;
     pthread_mutex_unlock(&output_buffer_mutex);
     usleep(1000); // attesa 1ms
    } while (count > 0);
 eof_flag = TRUE;
 // risvegli di thread sospesi in attesa di terminazione
 pthread_cond_signal(&input_buffer_not_empty);
 pthread_cond_signal(&output_buffer_not_full);
 pthread_cond_signal(&output_buffer_not_empty);
 pthread_exit(NULL);
}

void* process_thread(void* arg) // elaborazione dati
{
 while (!eof_flag)
      {
       // verifica/attesa presenza dati nel buffer di lettura
       pthread_mutex_lock(&input_buffer_mutex);
       if (input_buffer_count == 0) // input buffer vuoto
         pthread_cond_wait( &input_buffer_not_empty, &input_buffer_mutex);
       pthread_mutex_unlock(&input_buffer_mutex);
       // verifica/attesa disponbilita spazio per nuovi dati
       pthread_mutex_lock(&output_buffer_mutex);
       // output buffer pieno
       if (output_buffer_count >= BUFFER_BLOCKS)
         pthread_cond_wait( &output_buffer_not_full, &output_buffer_mutex);
       pthread_mutex_unlock(&output_buffer_mutex);
       if (eof_flag)
         break;
       if (crypt)
         DES_encrypt((BYTE*)&input_buffer[input_buffer_read_index*8],(BYTE*)&output_buffer[output_buffer_write_index*8],(BYTE*)password);
       else
           DES_decrypt((BYTE*)&input_buffer[input_buffer_read_index*8],(BYTE*)&output_buffer[output_buffer_write_index*8],(BYTE*)password);
       pthread_mutex_lock(&input_buffer_mutex);
       // incremento indice di lettura nel buffer di input
       input_buffer_read_index = (input_buffer_read_index + 1) % BUFFER_BLOCKS;
       // decremento contatore blocchi dati nel buffer di input
       input_buffer_count--;
       pthread_mutex_unlock(&input_buffer_mutex);
       // segnalazione condizione input buffer non pieno
       pthread_cond_signal(&input_buffer_not_full);
       pthread_mutex_lock(&output_buffer_mutex);
       // incremento indice di scrittura nel buffer di output
       output_buffer_write_index = (output_buffer_write_index + 1) % BUFFER_BLOCKS;
       // incremento contatore blocchi dati nel buffer di output
       output_buffer_count++;
       pthread_mutex_unlock(&output_buffer_mutex);
       // segnalazione condizione output buffer non vuoto
       pthread_cond_signal(&output_buffer_not_empty);
      }
 pthread_exit(NULL);
}

void* write_thread(void* arg) // scrittura su file
{
 while (!eof_flag)
      {
       // verifica/attesa presenza dati nel buffer di scrittura
       pthread_mutex_lock(&output_buffer_mutex);
       if (output_buffer_count == 0) // output buffer vuoto
         pthread_cond_wait( &output_buffer_not_empty, &output_buffer_mutex);
       pthread_mutex_unlock(&output_buffer_mutex);
       // scrittura di un blocco di 8 caratteri nel file
       fwrite(&output_buffer[output_buffer_read_index*8], 8, 1, destinazione);
       pthread_mutex_lock(&output_buffer_mutex);
       // incremento indice di lettura nel buffer di output
       output_buffer_read_index = (
       output_buffer_read_index + 1)% BUFFER_BLOCKS;
       // decremento contatore blocchi dati nel buffer di output
       output_buffer_count--;
       pthread_mutex_unlock(&output_buffer_mutex);
       // segnalazione condizione output buffer non pieno
       pthread_cond_signal(&output_buffer_not_full);
      }
 pthread_exit(NULL);
}


int main (int argc, char* argv[])
{
 int i, n;
 pthread_t R_thread, P_thread, W_thread;
 
 if (argc != 5)
   {
    printf( "Uso: %s file-origine file-destinazione password Cifra/Decifra\r\n",argv[0]);
    return 0;
   }
 // apertura in lettura file di origine
 origine = fopen(argv[1], "rb");
 if (origine == NULL)
   {
    printf("Errore apertura file \"%s\"\r\n",argv[1]);
    return 0;
   }
 // apertura in scrittura file di destinazione
 destinazione = fopen(argv[2], "wb");
 if (destinazione == NULL)
   {
    printf("Errore apertura file \"%s\"\r\n",argv[2]);
    fclose(origine);
    return 0;
   }
 // copia dei caratteri della password (massimo 8)
 n = strlen(argv[3]);
 if (n > 8)
   n = 8;
 for (i=0; i<n; i++)
    password[i] = argv[3][i];
 if (argv[4][0] == 'C')
   crypt = 1;
 else
     crypt = 0;
 input_buffer_count = 0;
 input_buffer_write_index = 0;
 input_buffer_read_index = 0;
 output_buffer_count = 0;
 output_buffer_write_index = 0;
 output_buffer_read_index = 0;
 eof_flag = FALSE;
 // creazione degli oggetti mutex per laccesso alle sezioni critiche
 pthread_mutex_init(&input_buffer_mutex, NULL);
 pthread_mutex_init (&output_buffer_mutex, NULL);
 // inizializzazione variabili condizione
 pthread_cond_init(&input_buffer_not_empty,NULL);
 pthread_cond_init(&input_buffer_not_full,NULL);
 pthread_cond_init(&output_buffer_not_empty,NULL);
 pthread_cond_init(&output_buffer_not_full,NULL);
 // creazione dei thread corrispondenti alle funzioni di lettura, elaborazione e scrittura
 pthread_create(&R_thread, NULL, &read_thread, NULL);
 pthread_create(&P_thread, NULL, &process_thread, NULL);
 pthread_create(&W_thread, NULL, &write_thread, NULL);
 // attesa della terminazione dei thread
 pthread_join(R_thread, NULL);
 pthread_join(P_thread, NULL);
 pthread_join(W_thread, NULL);
 // distruzione oggetti mutex per accesso sezioni critiche
 pthread_mutex_destroy(&input_buffer_mutex);
 pthread_mutex_destroy(&output_buffer_mutex);
 // distruzione variabili di condizione
 pthread_cond_destroy(&input_buffer_not_empty);
 pthread_cond_destroy(&input_buffer_not_full);
 pthread_cond_destroy(&output_buffer_not_empty);
 pthread_cond_destroy(&output_buffer_not_full);
 //chiusura file
 fclose(origine);
 fclose(destinazione);
 return 1;
}
