
#include <thread>
#include <chrono>
#include <mutex>
#include <ctime>
#include <iostream>

extern "C"
{
#include "meteo.h"
}

using namespace std;


struct METEO_DATA
{
	// timestamp
	unsigned short hour;
	unsigned short minute;
	unsigned short second;
	unsigned short day;
	unsigned short month;
	unsigned short year;
	// grandezze meteorologiche
	float temperature;
	float pressure;
	float relative_humidity;
	float wind_speed;
	float wind_direction;
};


class meteo_data_provider
{
private:

	bool active;
	mutex access;
	struct METEO_DATA data;
	thread* write_thread;
	unsigned int period;

	// thread scrittore produttore dei dati
	void write(void)
	{
		float t, p, rh, ws, wd;
		time_t now;
		struct tm *datetime;

		while (active)
		{
			this_thread::sleep_for(chrono::milliseconds(period));
			time(&now);
			datetime = localtime(&now);
			access.lock();
			data.day = datetime->tm_mday;
			data.month = datetime->tm_mon + 1;
			data.year = datetime->tm_year + 1900;
			data.hour = datetime->tm_hour;
			data.minute = datetime->tm_min;
			data.second = datetime->tm_sec;
			if (getActualTemperature(&t) >= 0)
				data.temperature = t;
			if (getActualPressure(&p))
				data.pressure = p;
			if (getActualRelativeHumidity(&rh))
				data.relative_humidity = rh;
			if (getActualWind(&ws, &wd) >= 0)
			{
				data.wind_speed = ws;
				data.wind_direction = wd;
			}
			access.unlock();
		}
	}

public:

	// costruttore
	meteo_data_provider(unsigned int period)
	{
		active = true;
		this->period = period;
		// creazione del thread
		write_thread = new thread(&meteo_data_provider::write, this);
	}

	~meteo_data_provider(void)
	{
		active = false;
		write_thread->join();
		delete write_thread;
	}

	// accesso thread-safe ai dati meteo attuali
	struct METEO_DATA getData(void)
	{
		struct METEO_DATA tmp;

		// acquisizione mutex di accesso ai dati
		access.lock();
		tmp = data;
		access.unlock();
		return tmp;
	}
};

meteo_data_provider *meteo;
bool active = true;

// thread lettore consumatore dei dtai
void leggi(unsigned int period)
{
	struct METEO_DATA data;

	while (active)
	{
		this_thread::sleep_for(chrono::seconds(period));
		data = meteo->getData();
		cout << this_thread::get_id() << "\t" << data.day << "/" << data.month << "/" << data.year << " " << data.hour << ":" << data.minute << ":" << data.second << "\t";
		cout << data.temperature << "C, " << data.pressure << "mbar, " << data.relative_humidity << "%, " << data.wind_speed << "Km/h, " << data.wind_direction << "" << endl;
	}
}

int main(int argc, char* argv[])
{
	thread* lettori[10];
	char tmp;
	
	// attivazione oggetto produttore dei dati (periodo di acquisizione: 1000ms)
	meteo = new meteo_data_provider(1000);
	// esecuzione dei thread lettori con periodo di accesso casuale
	for (int i = 0; i < 10; i++)
		lettori[i] = new thread(leggi, rand()%5+1);

	cout << "Premere un tasto qualsiasi seguito da \"Invio\" per terminare" << endl;
	cin >> tmp;
	active = false;
	// attesa della conclusione dei thread e chiusura dei file
	for (int i = 0; i < 10; i++)
	{
		lettori[i]->join();
		delete lettori[i];
	}
	delete meteo;

	return 0;
}
